@@ -2,6 +2,8 @@ import datetime
 import json
 import uuid
 import sys
+from typing import Tuple
+
 import dill
 import requests
 import pickle
@@ -12,7 +14,11 @@ import os
 import types
 import sklearn
 import wizata_dsapi
+from pandas import DataFrame
+from wizata_dsapi import MLModel
 
+import string
+import random
 
 from .plot import Plot
 from .request import Request
@@ -22,6 +28,7 @@ from .script import Script
 from .execution import Execution
 from .dsapi_json_encoder import DSAPIEncoder
 from .ds_dataframe import DSDataFrame
+from .model_toolkit import predict
 
 
 class WizataDSAPIClient:
@@ -162,12 +169,13 @@ class WizataDSAPIClient:
         else:
             raise TypeError("Type not supported.")
 
-    def get(self, obj=None, script_name=None):
+    def get(self, obj=None, script_name=None, experiment_key=None, model_key=None):
         """
         get full content of an object identified with is id.
 
         :param obj: object of a supported entity with at list its id
         :param script_name: alternatively you can use a function name to get a script
+        :param experiment_key: alternatively you can use an experiment key to get an experiment
         :return: object completed with all properties on server.
         """
         if script_name is not None:
@@ -180,7 +188,39 @@ class WizataDSAPIClient:
                 return dill.loads(script_bytes)
             else:
                 raise self.__raise_error(response)
-        if obj is not None:
+        elif experiment_key is not None:
+            response = requests.request("GET",
+                                        self.__url() + "experiments/",
+                                        params={
+                                            'key': experiment_key
+                                        },
+                                        headers=self.__header()
+                                        )
+            if response.status_code == 200:
+                obj = None
+                for obj in response.json():
+                    obj.from_json(response.json())
+                return obj
+            else:
+                raise self.__raise_error(response)
+
+        elif model_key is not None:
+            response = requests.request("GET",
+                                        self.__url() + "mlmodels/",
+                                        params={
+                                            'key': model_key
+                                        },
+                                        headers=self.__header()
+                                        )
+            if response.status_code == 200:
+                obj = None
+                for obj in response.json():
+                    obj.from_json(response.json())
+                return obj
+            else:
+                raise self.__raise_error(response)
+
+        elif obj is not None:
             if isinstance(obj, Execution):
                 response = requests.request("GET",
                                             self.__url() + "execute/" + str(obj.execution_id) + "/",
@@ -316,7 +356,7 @@ class WizataDSAPIClient:
     def upsert(self, obj):
         """
         upsert on object on the server
-        work with Script or directly your function using the technical name of the function
+        work with Script, MLModel or directly a function name
 
         :param obj: object to upsert on the server
         :return: ID of the object created or updated
@@ -330,9 +370,19 @@ class WizataDSAPIClient:
                                     headers=self.__header(),
                                     data=dill.dumps(obj))
             if response.status_code == 200:
+                obj.script_id = uuid.UUID(response.json()['id'])
                 return obj.script_id
             else:
                 raise self.__raise_error(response)
+        if isinstance(obj, MLModel):
+            response = requests.put(self.__url() + "mlmodels/",
+                                    headers=self.__header(),
+                                    data=pickle.dumps(obj))
+            if response.status_code == 200:
+                obj.model_id = uuid.UUID(response.json()['id'])
+                return obj.model_id
+            else:
+                raise self.__raise_error(response)
         else:
             raise TypeError("Type not supported.")
 
@@ -386,7 +436,8 @@ class WizataDSAPIClient:
               start: datetime,
               end: datetime,
               interval: int,
-              agg_method: str = "mean") -> pandas.DataFrame:
+              agg_method: str = "mean",
+              null: str = None) -> pandas.DataFrame:
         """
         Query a dataframe from API.
         :param agg_method:
@@ -394,6 +445,7 @@ class WizataDSAPIClient:
         :param start: start datetime of range to fetch
         :param end: end datetime of range to fetch
         :param interval: interval in milliseconds.
+        :param null: By default at 'drop' and dropping NaN values. If not intended behavior please set it to 'ignore' or 'all'.
         :return: dataframe
         """
         request = wizata_dsapi.Request()
@@ -407,6 +459,9 @@ class WizataDSAPIClient:
         if end is not None:
             request.end = end
 
+        if null is not None:
+            request.null = null
+
         request.set_aggregation(agg_method, interval)
 
         return self.get(request)
@@ -444,17 +499,17 @@ class WizataDSAPIClient:
             elif isinstance(dataframe, DSDataFrame):
                 execution.input_ds_dataframe = dataframe
         if script is not None:
-            if isinstance(script, uuid.UUID):
+            if isinstance(script, uuid.UUID) or (isinstance(script, str) and is_valid_uuid(script)):
                 execution.script = Script(script)
             elif isinstance(script, str):
                 execution.script = self.get(script_name=script)
             elif isinstance(script, Script):
                 execution.script = script
         if ml_model is not None:
-            if isinstance(ml_model, uuid.UUID):
+            if isinstance(ml_model, uuid.UUID) or (isinstance(script, str) and is_valid_uuid(ml_model)):
                 execution.ml_model = MLModel(ml_model)
             elif isinstance(ml_model, str):
-                execution.ml_model = MLModel(uuid.UUID(ml_model))
+                execution.ml_model = self.get(model_key=ml_model)
             elif isinstance(ml_model, MLModel):
                 execution.ml_model = ml_model
         if isAnomalyDetection:
@@ -487,7 +542,7 @@ class WizataDSAPIClient:
                 if "resultDataframe" in obj.keys() and obj["resultDataframe"]["id"] is not None:
                     execution.output_ds_dataframe = self.get(DSDataFrame(df_id=obj["resultDataframe"]["id"]))
                 if "anomaliesList" in obj.keys() and isinstance(obj["anomaliesList"], str):
-                    execution.anomalies = json.loads(execution["anomaliesList"])
+                    execution.anomalies = json.loads(obj["anomaliesList"])
                 return execution
             else:
                 raise self.__raise_error(response)
@@ -575,17 +630,17 @@ class WizataDSAPIClient:
             elif isinstance(dataframe, DSDataFrame):
                 execution.input_ds_dataframe = dataframe
         if script is not None:
-            if isinstance(script, uuid.UUID):
+            if isinstance(script, uuid.UUID) or (isinstance(script, str) and is_valid_uuid(script)):
                 execution.script = Script(script)
             elif isinstance(script, str):
                 execution.script = self.get(script_name=script)
             elif isinstance(script, Script):
                 execution.script = script
         if ml_model is not None:
-            if isinstance(ml_model, uuid.UUID):
+            if isinstance(ml_model, uuid.UUID) or (isinstance(script, str) and is_valid_uuid(ml_model)):
                 execution.ml_model = MLModel(ml_model)
             elif isinstance(ml_model, str):
-                execution.ml_model = MLModel(uuid.UUID(ml_model))
+                execution.ml_model = self.get(model_key=ml_model)
             elif isinstance(ml_model, MLModel):
                 execution.ml_model = ml_model
         if detect_anomalies:
@@ -634,6 +689,57 @@ class WizataDSAPIClient:
         else:
             raise KeyError('No valid arguments.')
 
+    def register_model(self,
+                       model_key,
+                       train_model,
+                       df: pandas.DataFrame,
+                       scaler=None,
+                       has_anomalies: bool = False,
+                       has_target_feat: bool = False,
+                       experiment_key = None) -> tuple[MLModel, pandas.DataFrame]:
+        """
+        Register a Machine Learning model to Wizata.
+        Model is tested by the API against a sample dataframe.
+        :param model_key: logical string id to identify the model.
+        :param train_model: trained model (must be compatible with pickle library)
+        :param df: sample dataframe.
+        :param scaler: scaler (must be compatible with pickle library)
+        :param has_anomalies: True is model generate Anomalies
+        :param has_target_feat: True if model need a target feature to be selected
+        :param experiment_key: Reference of an experiment to which link the generated ML Model
+        :return: registered ML Model , pandas.DataFrame
+        """
+        if model_key is None and isinstance(model_key, str):
+            raise ValueError('Please provide a str Model Key to identify the model.')
+        if train_model is None:
+            raise ValueError('Trained Machine Learning model should not be null')
+        elif df is None:
+            raise ValueError('A sample dataframe must be provided')
+
+        # Create a ML Model object
+        ml_model = wizata_dsapi.MLModel()
+        ml_model.trained_model = train_model
+        if scaler is not None:
+            ml_model.scaler = scaler
+        ml_model.has_anomalies = has_anomalies
+        ml_model.has_target_feat = has_target_feat
+        ml_model.input_columns = df.columns
+        ml_model.key = model_key
+
+        if experiment_key is not None:
+            ml_model.experimentId = self.get(experiment_key=experiment_key).experiment_id
+
+        try:
+            result_df = predict(df, ml_model)
+            if result_df is not None:
+                ml_model.status = "valid"
+                self.upsert(ml_model)
+                return ml_model, result_df
+            else:
+                raise RuntimeError('no dataframe was generated by your model while testing predict capabilities')
+        except Exception as e:
+            raise RuntimeError('not able to validated the model : ' + str(e))
+
 
 def api() -> WizataDSAPIClient:
     """
@@ -654,3 +760,11 @@ def api() -> WizataDSAPIClient:
         domain=os.environ.get('WIZATA_DOMAIN'),
         protocol=protocol
     )
+
+
+def is_valid_uuid(val):
+    try:
+        uuid.UUID(str(val))
+        return True
+    except ValueError:
+        return False
